#  Copyright (c) 2021, SBEL GPU Development Team
#  Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# This demo presents a cone penetrameter test with a soil sample made of clumped
# particles of various sizes. Before the test starts, when compress the terrain
# first, and note that the compressor used in this process has its position
# explicitly controlled step-by-step.
# =============================================================================

import DEME

import numpy as np
import os

if __name__ == "__main__":
    out_dir = "DemoOutput_ConePenetration/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("INFO")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetOutputContent(["ABSV"])
    DEMSim.SetMeshOutputFormat("VTK")
    DEMSim.SetContactOutputContent(["OWNER", "FORCE", "POINT"])

    # E, nu, CoR, mu, Crr... Material properties.
    mat_type_cone = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.8, "mu": 0.7, "Crr": 0.00})
    mat_type_terrain = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.8, "mu": 0.4, "Crr": 0.00})
    # If you don't have this line, then values will take average between 2 materials, when they are in contact
    DEMSim.SetMaterialPropertyPair("CoR", mat_type_cone, mat_type_terrain, 0.8)
    DEMSim.SetMaterialPropertyPair("mu", mat_type_cone, mat_type_terrain, 0.7)

    math_PI = 3.1415926
    cone_speed = 0.03
    step_size = 5e-6
    world_size = 2
    soil_bin_diameter = 0.584
    cone_surf_area = 323e-6
    cone_diameter = np.sqrt(cone_surf_area / math_PI) * 2
    DEMSim.InstructBoxDomainDimension(world_size, world_size, world_size)
    # No need to add simulation `world' boundaries, b/c we'll add a cylinderical container manually
    DEMSim.InstructBoxDomainBoundingBC("none", mat_type_terrain)
    # Now add a cylinderical boundary along with a bottom plane
    bottom = -0.5
    walls = DEMSim.AddExternalObject()
    walls.AddCylinder([0, 0, 0], [0, 0, 1],
                      soil_bin_diameter / 2., mat_type_terrain, 0)
    walls.AddPlane([0, 0, bottom], [0, 0, 1], mat_type_terrain)

    # Define the terrain particle templates
    # Calculate its mass and MOI
    terrain_density = 2.6e3
    clump_vol = 5.5886717
    mass = terrain_density * clump_vol
    MOI = np.array([2.928, 2.6029, 3.9908]) * terrain_density
    # Then load it to system
    my_template = DEMSim.LoadClumpType(mass, MOI.tolist(
    ), DEME.GetDEMEDataFile("clumps/3_clump.csv"), mat_type_terrain)
    my_template.SetVolume(clump_vol)
    # Decide the scalings of the templates we just created (so that they are... like particles, not rocks)
    scale = 0.0044
    my_template.Scale(scale)

    # Sampler to sample
    sampler = DEME.HCPSampler(scale * 3.)
    fill_height = 0.5
    fill_center = [0, 0, bottom + fill_height / 2]
    fill_radius = soil_bin_diameter / 2. - scale * 3.
    input_xyz = sampler.SampleCylinderZ(
        fill_center, fill_radius, fill_height / 2 - scale * 2.)
    DEMSim.AddClumps(my_template, input_xyz)
    print(f"Total num of particles: {len(input_xyz)}")

    # Load in the cone used for this penetration test
    cone_tip = DEMSim.AddWavefrontMeshObject(
        DEME.GetDEMEDataFile("mesh/cone.obj"), mat_type_cone)
    cone_body = DEMSim.AddWavefrontMeshObject(
        DEME.GetDEMEDataFile("mesh/cyl_r1_h2.obj"), mat_type_cone)
    print(
        f"Total num of triangles: {cone_tip.GetNumTriangles() + cone_body.GetNumTriangles()}")

    # The initial cone mesh has base radius 1, and height 1. Let's stretch it a bit so it has a 60deg tip, instead of
    # 90deg.
    tip_height = np.sqrt(3.)
    cone_tip.Scale([1, 1, tip_height])
    # Then set mass properties
    cone_mass = 7.8e3 * tip_height / 3 * math_PI
    cone_tip.SetMass(cone_mass)
    # You can checkout https://en.wikipedia.org/wiki/List_of_moments_of_inertia
    cone_tip.SetMOI([cone_mass * (3. / 20. + 3. / 80. * tip_height * tip_height),
                    cone_mass * (3. / 20. + 3. / 80. * tip_height * tip_height), 3 * cone_mass / 10])
    # This cone mesh has its tip at the origin. And, float4 quaternion pattern is (x, y, z, w).
    cone_tip.InformCentroidPrincipal(
        [0, 0, 3. / 4. * tip_height], [0, 0, 0, 1])
    # Note the scale method will scale mass and MOI automatically. But this only goes for the case you scale xyz all
    # together; otherwise, the MOI scaling will not be accurate and you should manually reset them.
    cone_tip.Scale(cone_diameter / 2)
    cone_tip.SetFamily(2)

    # The define the body that is connected to the tip
    body_mass = 7.8e3 * math_PI
    cone_body.SetMass(body_mass)
    cone_body.SetMOI([body_mass * 7 / 12, body_mass * 7 / 12, body_mass / 2])
    # This cyl mesh (h = 2m, r = 1m) has its center at the origin. So the following call actually has no effect...
    cone_body.InformCentroidPrincipal([0, 0, 0], [0, 0, 0, 1])
    cone_body.Scale([cone_diameter / 2, cone_diameter / 2, 0.5])
    cone_body.SetFamily(2)

    # Track the cone_tip
    tip_tracker = DEMSim.Track(cone_tip)
    body_tracker = DEMSim.Track(cone_body)

    # Because the cone's motion is completely pre-determined, we can just prescribe family 1
    DEMSim.SetFamilyPrescribedLinVel(1, "0", "0", f"-{cone_speed}")
    # Cone is initially in family 2, sleeping...
    DEMSim.SetFamilyFixed(2)
    DEMSim.DisableContactBetweenFamilies(0, 2)

    # Now add a plane to compress the sample
    compressor = DEMSim.AddExternalObject()
    compressor.AddPlane([0, 0, 0], [0, 0, -1], mat_type_terrain)
    compressor.SetFamily(10)
    DEMSim.SetFamilyFixed(10)
    compressor_tracker = DEMSim.Track(compressor)

    # Some inspectors
    max_z_finder = DEMSim.CreateInspector("clump_max_z")
    total_mass_finder = DEMSim.CreateInspector("clump_mass")

    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])
    # CD freq will be auto-adapted so it does not matter much here.
    DEMSim.SetCDUpdateFreq(20)
    # Max velocity info is generally just for the solver's reference and the user do not have to set it. The solver
    # wouldn't take into account a vel larger than this when doing async-ed contact detection: but this vel won't
    # happen anyway and if it does, something already went wrong.
    DEMSim.SetMaxVelocity(10.)
    DEMSim.Initialize()

    # Settle
    DEMSim.DoDynamicsThenSync(0.8)

    # Compress until dense enough
    currframe = 0
    curr_step = 0
    fps = 20
    out_steps = (int)(1.0 / (fps * step_size))
    compressor_vel = 0.05
    terrain_max_z = max_z_finder.GetValue()
    init_max_z = terrain_max_z
    bulk_density = -10000.
    while (bulk_density < 1500.):
        matter_mass = total_mass_finder.GetValue()
        total_volume = math_PI * \
            (soil_bin_diameter * soil_bin_diameter / 4) * (terrain_max_z - bottom)
        bulk_density = matter_mass / total_volume
        if (curr_step % out_steps == 0):
            print(f"Outputting frame: {currframe}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{currframe:04d}.csv")
            # meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
            DEMSim.WriteSphereFile(filename)
            # DEMSim.WriteMeshFile(meshname)
            currframe += 1
            print(f"Compression bulk density: {bulk_density}", flush=True)

        terrain_max_z -= compressor_vel * step_size
        compressor_tracker.SetPos([0, 0, terrain_max_z])
        DEMSim.DoDynamics(step_size)
        curr_step += 1

    # Then gradually remove the compressor
    while (terrain_max_z < init_max_z):
        if (curr_step % out_steps == 0):
            print(f"Outputting frame: {currframe}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{currframe:04d}.csv")
            # meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
            DEMSim.WriteSphereFile(filename)
            # DEMSim.WriteMeshFile(meshname)
            matter_mass = total_mass_finder.GetValue()
            total_volume = math_PI * \
                (soil_bin_diameter * soil_bin_diameter / 4) * \
                (max_z_finder.GetValue() - bottom)
            bulk_density = matter_mass / total_volume
            print(f"Compression bulk density: {bulk_density}", flush=True)
            currframe += 1

        terrain_max_z += compressor_vel * step_size
        compressor_tracker.SetPos([0, 0, terrain_max_z])
        DEMSim.DoDynamics(step_size)
        curr_step += 1

    # Remove compressor
    DEMSim.DoDynamicsThenSync(0.)
    DEMSim.DisableContactBetweenFamilies(0, 10)
    DEMSim.DoDynamicsThenSync(0.2)
    terrain_max_z = max_z_finder.GetValue()

    sim_end = 7.0
    fps = 2500
    frame_time = 1.0 / fps
    print(f"Output at {fps} FPS")

    # Put the cone in place
    starting_height = terrain_max_z + 0.03
    # Its initial position should be right above the cone tip...
    body_tracker.SetPos(
        [0, 0, 0.5 + (cone_diameter / 2 / 4 * tip_height) + starting_height])
    # Note that position of objects is always the location of their centroid
    tip_tracker.SetPos([0, 0, starting_height])
    # The tip location, used to measure penetration length
    tip_z = -cone_diameter / 2 * 3 / 4 * tip_height + starting_height

    # Enable cone
    DEMSim.ChangeFamily(2, 1)
    matter_mass = total_mass_finder.GetValue()
    total_volume = math_PI * (soil_bin_diameter *
                              soil_bin_diameter / 4) * (terrain_max_z - bottom)
    bulk_density = matter_mass / total_volume
    print(f"Bulk density: {bulk_density}", flush=True)

    tip_z_when_first_hit = 9999.
    hit_terrain = False
    frame_count = 0

    t = 0.
    while (t < sim_end):
        # terrain_max_z = max_z_finder.GetValue()
        forces = tip_tracker.ContactAcc()
        # Note cone_mass is not the true mass, b/c we scaled the the cone tip! So we use true mass by using cone_mass
        forces = np.array(forces) * cone_tip.Mass()
        # forces[2] is the z dir force
        pressure = np.abs(forces[2]) / cone_surf_area
        if ((pressure > 1e-4) and not (hit_terrain)):
            hit_terrain = True
            tip_z_when_first_hit = tip_z

        penetration = tip_z_when_first_hit - tip_z if hit_terrain else 0.
        print(f"Time: {t}", flush=True)
        print(f"Z coord of tip: {tip_z}", flush=True)
        print(f"Penetration: {penetration}", flush=True)
        print(
            f"Force on cone: {forces[0]}, {forces[1]}, {forces[2]}", flush=True)
        print(f"Pressure: {pressure}", flush=True)

        if (frame_count % 500 == 0):
            print(f"Outputting frame: {currframe}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{currframe:04d}.csv")
            meshname = os.path.join(
                out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
            DEMSim.WriteSphereFile(filename)
            DEMSim.WriteMeshFile(meshname)
            DEMSim.ShowThreadCollaborationStats()
            currframe += 1

        DEMSim.DoDynamicsThenSync(frame_time)
        tip_z -= cone_speed * frame_time

        frame_count += 1
        t += frame_time

    DEMSim.ShowMemStats()
    print(f"ConePenetration demo exiting...")



# Copyright (c) 2021, SBEL GPU Development Team
# Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# This demo features a mesh-represented bladed mixer interacting with clump-represented
# DEM particles.
# =============================================================================

import DEME
from DEME import HCPSampler

import numpy as np
import os
import time

if __name__ == "__main__":
    out_dir = "DemoOutput_Mixer/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("STEP_METRIC")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetOutputContent(["ABSV", "XYZ"])
    DEMSim.SetMeshOutputFormat("VTK")

    # If you don't need individual force information, then this option makes the solver run a bit faster.
    DEMSim.SetNoForceRecord(True)

    # E, nu, CoR, mu, Crr... Material properties
    mat_type_mixer = DEMSim.LoadMaterial(
        {"E": 1e8, "nu":  0.3, "CoR":  0.6, "mu":  0.5, "Crr": 0.0})
    mat_type_granular = DEMSim.LoadMaterial(
        {"E":  1e8, "nu":  0.3, "CoR":  0.8, "mu":  0.2, "Crr":  0.0})
    # If you don't have this line, then mu between mixer material and granular material will be 0.35 (average of the two).
    DEMSim.SetMaterialPropertyPair(
        "CoR", mat_type_mixer, mat_type_granular, 0.5)

    step_size = 5e-6
    world_size = 1
    chamber_height = world_size / 3.
    fill_height = chamber_height
    chamber_bottom = -world_size / 2.
    fill_bottom = chamber_bottom + chamber_height

    DEMSim.InstructBoxDomainDimension(world_size, world_size, world_size)
    DEMSim.InstructBoxDomainBoundingBC("all", mat_type_granular)

    # Now add a cylinderical boundary
    walls = DEMSim.AddExternalObject()
    walls.AddCylinder([0, 0, 0], [0, 0, 1], world_size / 2., mat_type_mixer, 0)

    mixer = DEMSim.AddWavefrontMeshObject(
        DEME.GetDEMEDataFile("mesh/internal_mixer.obj"), mat_type_mixer)
    print(f"Total num of triangles: {mixer.GetNumTriangles()}")
    mixer.Scale([world_size / 2, world_size / 2, chamber_height])
    mixer.SetFamily(10)
    # Define the prescribed motion of mixer
    DEMSim.SetFamilyPrescribedAngVel(10, "0", "0", "3.14159")

    granular_rad = 0.005
    # Calculate its mass and MOI
    mass = 2.6e3 * 5.5886717
    MOI = np.array([2.928, 2.6029, 3.9908]) * 2.6e3
    template_granular = DEMSim.LoadClumpType(mass, MOI.tolist(
    ), DEME.GetDEMEDataFile("clumps/3_clump.csv"), mat_type_granular)
    template_granular.Scale(granular_rad)

    # Track the mixer
    mixer_tracker = DEMSim.Track(mixer)

    # Sampler uses hex close-packing
    sampler = HCPSampler(3.0 * granular_rad)
    fill_center = [0, 0, fill_bottom + fill_height / 2]
    fill_radius = world_size / 2. - 2. * granular_rad
    input_xyz = sampler.SampleCylinderZ(
        fill_center, fill_radius, fill_height / 2)
    DEMSim.AddClumps(template_granular, input_xyz)
    print(f"Total num of particles: {len(input_xyz)}")

    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])
    DEMSim.SetCDUpdateFreq(40)
    # Mixer has a big angular velocity-contributed linear speed at its blades, this is something the solver do not
    # account for, for now. And that means it needs to be added as an estimated value.
    DEMSim.SetExpandSafetyAdder(2.0)
    # You usually don't have to worry about initial bin size. In very rare cases, init bin size is so bad that auto bin
    # size adaption is effectless, and you should notice in that case kT runs extremely slow. Then in that case setting
    # init bin size may save the simulation.
    # DEMSim.SetInitBinSize(25 * granular_rad);
    DEMSim.SetCDNumStepsMaxDriftMultipleOfAvg(1.2)
    DEMSim.SetCDNumStepsMaxDriftAheadOfAvg(6)
    DEMSim.SetSortContactPairs(True)
    # DEMSim.DisableAdaptiveBinSize();
    DEMSim.SetErrorOutVelocity(20.)
    # Force the solver to error out if something went crazy. A good practice to add them, but not necessary.
    DEMSim.SetErrorOutAvgContacts(50)
    DEMSim.Initialize()

    sim_end = 10.0
    fps = 20
    frame_time = 1.0 / fps

    # Keep tab of the max velocity in simulation
    max_v_finder = DEMSim.CreateInspector("clump_max_absv")

    print(f"Output at {fps} FPS")
    currframe = 0

    mixer_tracker.SetPos([0, 0, chamber_bottom + chamber_height / 2.0])

    t = 0.
    start = time.process_time()
    while (t < sim_end):
        print(f"Frame: {currframe}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{currframe:04d}.csv")
        meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        currframe += 1

        max_v = max_v_finder.GetValue()
        print(
            f"Max velocity of any point in simulation is {max_v}", flush=True)
        print(
            f"Solver's current update frequency (auto-adapted): {DEMSim.GetUpdateFreq()}", flush=True)
        print(
            f"Average contacts each sphere has: {DEMSim.GetAvgSphContacts()}", flush=True)

        mixer_moi = np.array(mixer_tracker.MOI())
        mixer_acc = np.array(mixer_tracker.ContactAngAccLocal())
        mixer_torque = mixer_acc * mixer_moi
        print(
            f"Contact torque on the mixer is {mixer_torque[0]}, {mixer_torque[1]}, {mixer_torque[2]}", flush=True)

        DEMSim.DoDynamics(frame_time)
        DEMSim.ShowThreadCollaborationStats()

        t += frame_time

    elapsed_time = time.process_time() - start
    print(f"{elapsed_time} seconds (wall time) to finish this simulation")

    DEMSim.ShowTimingStats()
    DEMSim.ShowMemStats()
    print("Mixer demo exiting...")



#  Copyright (c) 2021, SBEL GPU Development Team
#  Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# This demo lets a rod with some electric charges stick into a pile of granular
# material that is also charged. The electrostatic force shows its effect. The
# electric charges are even moving during the simulation. This is done through
# `Geometry Wildcard', where the amount of charges is associated with each
# sphere component (of clumps) and triangles (of meshes). Then a custom force
# model is used to derive the electrostatic force in addition to contact forces.
#
# NOTE: If you want to create your own force model, it's probably a good idea
# to understand the default model in the file FullHertzianForceModel.cu
# (Hertzian--Mindlin), how the normal and tangential forces are calculated,
# then use it as a starting point to add your own force, like in this demo.
# =============================================================================

import DEME
from DEME import HCPSampler

import numpy as np
import os

math_PI = 3.1415927

# For clarity, we attach the force model file here.
# This model can be found in file ForceModelWithElectrostatic.cu, too.


def force_model():
    model = """
        /////////////////////////////////////////////////////////////
        // The first part is just the standard full Hertzian--Mindlin
        /////////////////////////////////////////////////////////////

        // No need to do any contact force calculation if no contact. And it can happen,
        // since we added extra contact margin for adding electrostatic forces before 
        // physical contacts emerge.
        if (overlapDepth > 0) {
            // Material properties
            float E_cnt, G_cnt, CoR_cnt, mu_cnt, Crr_cnt;
            {
                // E and nu are associated with each material, so obtain them this way
                float E_A = E[bodyAMatType];
                float nu_A = nu[bodyAMatType];
                float E_B = E[bodyBMatType];
                float nu_B = nu[bodyBMatType];
                matProxy2ContactParam<float>(E_cnt, G_cnt, E_A, nu_A, E_B, nu_B);
                // CoR, mu and Crr are pair-wise, so obtain them this way
                CoR_cnt = CoR[bodyAMatType][bodyBMatType];
                mu_cnt = mu[bodyAMatType][bodyBMatType];
                Crr_cnt = Crr[bodyAMatType][bodyBMatType];
            }

            float3 rotVelCPA, rotVelCPB;
            {
                // We also need the relative velocity between A and B in global frame to use in the damping terms
                // To get that, we need contact points' rotational velocity in GLOBAL frame
                // This is local rotational velocity (the portion of linear vel contributed by rotation)
                rotVelCPA = cross(ARotVel, locCPA);
                rotVelCPB = cross(BRotVel, locCPB);
                // This is mapping from local rotational velocity to global
                applyOriQToVector3<float, deme::oriQ_t>(rotVelCPA.x, rotVelCPA.y, rotVelCPA.z, AOriQ.w, AOriQ.x, AOriQ.y, AOriQ.z);
                applyOriQToVector3<float, deme::oriQ_t>(rotVelCPB.x, rotVelCPB.y, rotVelCPB.z, BOriQ.w, BOriQ.x, BOriQ.y, BOriQ.z);
            }

            // A few re-usables
            float mass_eff, sqrt_Rd, beta;
            float3 vrel_tan;
            float3 delta_tan = make_float3(delta_tan_x, delta_tan_y, delta_tan_z);

            // Normal force part
            {
                // The (total) relative linear velocity of A relative to B
                const float3 velB2A = (ALinVel + rotVelCPA) - (BLinVel + rotVelCPB);
                const float projection = dot(velB2A, B2A);
                vrel_tan = velB2A - projection * B2A;

                // Now we already have sufficient info to update contact history
                {
                    delta_tan += ts * vrel_tan;
                    const float disp_proj = dot(delta_tan, B2A);
                    delta_tan -= disp_proj * B2A;
                    delta_time += ts;
                }

                mass_eff = (AOwnerMass * BOwnerMass) / (AOwnerMass + BOwnerMass);
                sqrt_Rd = sqrt(overlapDepth * (ARadius * BRadius) / (ARadius + BRadius));
                const float Sn = 2. * E_cnt * sqrt_Rd;

                const float loge = (CoR_cnt < DEME_TINY_FLOAT) ? log(DEME_TINY_FLOAT) : log(CoR_cnt);
                beta = loge / sqrt(loge * loge + deme::PI_SQUARED);

                const float k_n = deme::TWO_OVER_THREE * Sn;
                const float gamma_n = deme::TWO_TIMES_SQRT_FIVE_OVER_SIX * beta * sqrt(Sn * mass_eff);

                force += (k_n * overlapDepth + gamma_n * projection) * B2A;
            }

            // Rolling resistance part
            if (Crr_cnt > 0.0) {
                // Figure out if we should apply rolling resistance force
                bool should_add_rolling_resistance = true;
                {
                    const float R_eff = sqrtf((ARadius * BRadius) / (ARadius + BRadius));
                    const float kn_simple = deme::FOUR_OVER_THREE * E_cnt * sqrtf(R_eff);
                    const float gn_simple = -2.f * sqrtf(deme::FIVE_OVER_THREE * mass_eff * E_cnt) * beta * powf(R_eff, 0.25f);

                    const float d_coeff = gn_simple / (2.f * sqrtf(kn_simple * mass_eff));

                    if (d_coeff < 1.0) {
                        float t_collision = deme::PI * sqrtf(mass_eff / (kn_simple * (1.f - d_coeff * d_coeff)));
                        if (delta_time <= t_collision) {
                            should_add_rolling_resistance = false;
                        }
                    }
                }
                // If should, then compute it (using Schwartz model)
                if (should_add_rolling_resistance) {
                    // Tangential velocity (only rolling contribution) of B relative to A, at contact point, in global
                    const float3 v_rot = rotVelCPB - rotVelCPA;
                    // This v_rot is only used for identifying resistance direction
                    const float v_rot_mag = length(v_rot);
                    if (v_rot_mag > DEME_TINY_FLOAT) {
                        // You should know that Crr_cnt * normal_force is the underlying formula, and in our model,
                        // it is a `force' that produces torque only, instead of also cancelling out friction.
                        // Its direction is that it `resists' rotation, see picture in
                        // https://en.wikipedia.org/wiki/Rolling_resistance.
                        torque_only_force = (v_rot / v_rot_mag) * (Crr_cnt * length(force));
                    }
                }
            }

            // Tangential force part
            if (mu_cnt > 0.0) {
                const float kt = 8. * G_cnt * sqrt_Rd;
                const float gt = -deme::TWO_TIMES_SQRT_FIVE_OVER_SIX * beta * sqrt(mass_eff * kt);
                float3 tangent_force = -kt * delta_tan - gt * vrel_tan;
                const float ft = length(tangent_force);
                if (ft > DEME_TINY_FLOAT) {
                    // Reverse-engineer to get tangential displacement
                    const float ft_max = length(force) * mu_cnt;
                    if (ft > ft_max) {
                        tangent_force = (ft_max / ft) * tangent_force;
                        delta_tan = (tangent_force + gt * vrel_tan) / (-kt);
                    }
                } else {
                    tangent_force = make_float3(0, 0, 0);
                }
                // Use force to collect tangent_force
                force += tangent_force;
            }

            // Finally, make sure we update those wildcards (in this case, contact history)
            delta_tan_x = delta_tan.x;
            delta_tan_y = delta_tan.y;
            delta_tan_z = delta_tan.z;
        } else { 
            // This is to be more rigorous. If in fact no physical contact, then contact wildcards (such as contact history)
            // should be cleared (they will also be automatically cleared if the contact is no longer detected).
            delta_time = 0;
            delta_tan_x = 0;
            delta_tan_y = 0;
            delta_tan_z = 0;
        }

        /////////////////////////////////////////////////
        // Now we add an extra electrostatic force
        ////////////////////////////////////////////////
        {
            const float k = 8.99e9;
            const double ABdist2 = dot(bodyAPos - bodyBPos, bodyAPos - bodyBPos);
            // If Q_A and Q_B are the same sign, then the force pushes A away from B, so B2A is the direction.
            force += k * Q_A[AGeo] * Q_B[BGeo] / ABdist2 * (B2A);
            // Fun part: we can modify the electric charge on the fly. But we have to use atomic, since multiple contacts
            // can modify the same Q.
            // But this is not recommend unless you understand what you are doing, and there are a lot of details related to it.
            // For example, although the charges transfer between geometries, the geometries within one clump cannot
            // re-distribute elec charges among them, since no contact among geometries in one clump. Still, you could write
            // your own subroutine to further modify those geometry and/or own wildcards in your script, or within the force
            // model. 
            // On the other hand, if you do not need to modify the wildcards, you just need to use them for calculating
            // the force, then that is probably easier and with less strings attached to it. I can see this being more
            // useful.
            if (overlapDepth > 0) {  // Exchange the elec charge only in physical contact
                float avg_Q = (Q_A[AGeo] + Q_B[BGeo]) / 2.;
                float A_change_dir = (abs(avg_Q - Q_A[AGeo]) > 1e-11) ? (avg_Q - Q_A[AGeo]) / abs(avg_Q - Q_A[AGeo]) : 0.;
                // Modify the charge they carry... the rate is 1e-8 per second
                atomicAdd(Q_A + AGeo, A_change_dir * 1e-8 * ts);
                atomicAdd(Q_B + BGeo, -A_change_dir * 1e-8 * ts);
            }
        }
    """

    return model


if __name__ == "__main__":
    out_dir = "DemoOutput_Electrostatic/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("INFO")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetOutputContent(["ABSV", "XYZ"])
    DEMSim.SetMeshOutputFormat("VTK")
    DEMSim.SetContactOutputContent(["OWNER", "FORCE", "POINT"])

    # If you don't need individual force information, then this option makes the solver run a bit faster.
    DEMSim.SetNoForceRecord(True)

    # E, nu, CoR, mu, Crr...
    mat_type_rod = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.5, "mu": 0.7, "Crr": 0.00})
    mat_type_terrain = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.5, "mu": 0.4, "Crr": 0.00})
    # If you don't have this line, then values will take average between 2 materials, when they are in contact
    DEMSim.SetMaterialPropertyPair("CoR", mat_type_rod, mat_type_terrain, 0.8)
    DEMSim.SetMaterialPropertyPair("mu", mat_type_rod, mat_type_terrain, 0.7)
    # We can specify the force model using a string.
    # This force model is the standard Hertzian--Mindlin model, plus an electrostatic force, based on something
    # called a `geometry wildcard'. This is an extra property that we can associate with each geometry entity,
    # such as triangle and sphere. We use this value to derive the electrostatic force. But first, we need to
    # declare in the force model that a geometry wildcard is in use...
    my_force_model = DEMSim.DefineContactForceModel(force_model())
    # my_force_model = DEMSim.ReadContactForceModel("ForceModelWithElectrostatic.cu");

    # Those following lines are needed. We must let the solver know that those var names are history variable etc.
    my_force_model.SetMustHaveMatProp(set(["E", "nu", "CoR", "mu", "Crr"]))
    my_force_model.SetMustPairwiseMatProp(set(["CoR", "mu", "Crr"]))
    my_force_model.SetPerContactWildcards(
        set(["delta_time", "delta_tan_x", "delta_tan_y", "delta_tan_z"]))
    # Use variable name `Q' for the amount of electrc charge.
    # NOTE! If you call it Q here, then you can refer to this wildcard array using variable names Q_A amd Q_B in
    # your custom force model.
    my_force_model.SetPerGeometryWildcards(set(["Q"]))

    init_charge = 2e-8  # Coulomb as the unit...
    cone_speed = 0.1
    step_size = 5e-6
    world_size = 2
    soil_bin_diameter = 0.584
    rod_length = 0.5
    rod_surf_area = 323e-6
    rod_diameter = np.sqrt(rod_surf_area / math_PI) * 2
    DEMSim.InstructBoxDomainDimension(world_size, world_size, world_size)
    # No need to add simulation `world' boundaries, b/c we'll add a cylinderical container manually
    DEMSim.InstructBoxDomainBoundingBC("none", mat_type_terrain)
    # Now add a cylinderical boundary along with a bottom plane
    bottom = -0.5
    walls = DEMSim.AddExternalObject()
    walls.AddCylinder([0, 0, 0], [0, 0, 1],
                      soil_bin_diameter / 2., mat_type_terrain, 0)
    walls.AddPlane([0, 0, bottom], [0, 0, 1], mat_type_terrain)
    walls.AddPlane([0, 0, world_size / 2. - world_size / 20.],
                   [0, 0, -1], mat_type_terrain)

    # Define the terrain particle templates
    # Calculate its mass and MOI
    terrain_density = 2.6e3
    clump_vol = 4. / 3. * math_PI
    mass = terrain_density * clump_vol
    MOI = np.array([2.0 / 5.0, 2.0 / 5.0, 2.0 / 5.0]) * mass
    # Then load it to system
    my_template = DEMSim.LoadClumpType(mass, MOI.tolist(), DEME.GetDEMEDataFile(
        "clumps/spiky_sphere.csv"), mat_type_terrain)
    my_template.SetVolume(clump_vol)
    # Decide the scalings of the templates we just created (so that they are... like particles, not rocks)
    scale = 0.01
    my_template.Scale(scale)

    # Sampler to sample
    sampler = HCPSampler(scale * 2.4)
    fill_height = 1.
    fill_center = [0, 0, bottom + fill_height / 2]
    fill_radius = soil_bin_diameter / 2. - scale * 3.

    input_xyz = sampler.SampleCylinderZ(
        fill_center, fill_radius, fill_height / 2 - scale * 2.)

    particles = DEMSim.AddClumps(my_template, input_xyz)

    print(f"Total num of particles: {particles.GetNumClumps()}")
    print(f"Total num of spheres: {particles.GetNumSpheres()}")
    # Add electric charge Q to each sphere. It is important to add it to each sphere, not each clump. This
    # is because the contact pairs are resolved between geometries, not clumps. Using clumps leads to
    # double-count or triple-count or...
    particles.AddGeometryWildcard(
        "Q", [init_charge] * particles.GetNumSpheres())

    # Load in the cone used for this penetration test
    rod_body = DEMSim.AddWavefrontMeshObject(
        DEME.GetDEMEDataFile("mesh/cyl_r1_h2.obj"), mat_type_rod, True, False)
    num_tri = rod_body.GetNumTriangles()
    print(f"Total num of triangles: {num_tri}")

    # The define the properties of the rod
    body_mass = 7.8e3 * math_PI
    rod_body.SetMass(body_mass)
    rod_body.SetMOI([body_mass * 7 / 12, body_mass * 7 / 12, body_mass / 2])
    # This cyl mesh (h = 2m, r = 1m) has its center at the origin. So the following call actually has no effect...
    rod_body.InformCentroidPrincipal([0, 0, 0], [0, 0, 0, 1])
    rod_body.Scale([rod_diameter / 2., rod_diameter / 2., rod_length / 2.])
    rod_body.SetFamily(1)
    # Just fix this rod: We will manually impose its motion later.
    DEMSim.SetFamilyFixed(1)
    # We can set the geometry wildcard Q here. But we'll instead show how to modify that using a tracker.
    # rod_body.AddGeometryWildcard("Q", [init_charge] * num_tri)

    # Track the rod
    rod_tracker = DEMSim.Track(rod_body)

    # Some inspectors
    max_z_finder = DEMSim.CreateInspector("clump_max_z")

    # This ensures that the force model is in effect even for some `contacts' that are not physical contacts, since the
    # electrostatic force should work in a distance. We assume that this effect won't be longer than 5mm.
    DEMSim.SetFamilyExtraMargin(0, 0.005)
    DEMSim.SetFamilyExtraMargin(1, 0.005)
    # If you know your extra margin policy gonna make the average number of contacts per sphere huge, then set this so
    # that the solver does not error out when checking it.
    DEMSim.SetErrorOutAvgContacts(200)

    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])
    DEMSim.Initialize()

    sim_end = 9.0
    fps = 20
    frame_time = 1.0 / fps
    print(f"Output at {fps} FPS")
    out_steps = (int)(1.0 / (fps * step_size))
    currframe = 0
    step_count = 0

    # Settle
    DEMSim.DisableContactBetweenFamilies(0, 1)

    t = 0.
    while (t < 1.):
        print(f"Outputting frame: {currframe}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{currframe:04d}.csv")
        meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        DEMSim.ShowThreadCollaborationStats()
        currframe += 1

        DEMSim.DoDynamics(frame_time)
        t += frame_time

    # Put the cone in place
    terrain_max_z = max_z_finder.GetValue()
    current_height = terrain_max_z + 0.03
    # Its initial position should be right above the granular material (but accounting for the fact that the coordinate
    # system center of the rod is in its middle)
    rod_tracker.SetPos([0, 0, rod_length / 2. + current_height])

    DEMSim.EnableContactBetweenFamilies(0, 1)

    # We demonstrate using trackers to set a geometry wildcard. Q is now set for each triangle facet, and it's
    # the opposite charge to the particles. So the rod should attract the particles.
    rod_tracker.SetGeometryWildcardValues("Q", [-10. * init_charge] * num_tri)

    t = 0.
    while (t < sim_end):
        if (step_count % out_steps == 0):
            print(f"Outputting frame: {currframe}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{currframe:04d}.csv")
            meshname = os.path.join(
                out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
            DEMSim.WriteSphereFile(filename)
            DEMSim.WriteMeshFile(meshname)
            DEMSim.ShowThreadCollaborationStats()
            currframe += 1

        # Means advance simulation by one time step
        DEMSim.DoStepDynamics()
        step_count += 1
        t += step_size
        # The rod first goes down into the material, then goes up, then stay in place to let us watch how the particles
        # affected by the electrostatic force move.
        if (t < 1. / 3. * sim_end):
            current_height -= cone_speed * step_size
        elif (t < 2. / 3. * sim_end):
            current_height += cone_speed * step_size
        # else the rod does not move
        rod_tracker.SetPos([0, 0, rod_length / 2. + current_height])

    DEMSim.ShowTimingStats()
    DEMSim.ShowMemStats()
    print("Electrostatic demo exiting...")



#  Copyright (c) 2021, SBEL GPU Development Team
#  Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# This demo reproduces the Game of Life in DEME simulator, to showcase the
# flexibility of its APIs.
# =============================================================================

import DEME

import numpy as np
import os
import random

if __name__ == "__main__":
    out_dir = "DemoOutput_GameOfLife/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("INFO")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetOutputContent(["XYZ"])
    DEMSim.SetMeshOutputFormat("VTK")
    DEMSim.EnsureKernelErrMsgLineNum(True)

    grid_size = 1.0
    r = (1.45 * grid_size) / 2.0
    world_size = 2500.0
    n_init = 10000

    # Material is formaility... you can opt not to set it at all, it works the same
    mat_type_1 = DEMSim.LoadMaterial({"junk": 1.0})

    template_sphere = DEMSim.LoadSphereType(1.0, r, mat_type_1)

    input_template_num = []
    family_code = []

    # Generate a flat layer of particles in this domain
    input_xyz = DEME.DEMBoxGridSampler(
        [0, 0, 0], [world_size / 2.0, world_size / 2.0, 0.001], grid_size)
    num_cells = len(input_xyz)
    # Number cells per row
    num_row = round(np.sqrt(num_cells))

    # Use code 10 to represent dead cells, and don't output dead cells
    family_code += [10] * len(input_xyz)
    DEMSim.DisableFamilyOutput(10)

    # Init patterns adding...
    for i in range(n_init):
        offset = 0

        while (offset < 5 * num_row) or (offset > (num_row - 5) * num_row):
            offset = random.randint(0, num_cells - 1)
        # Mark some of them to be alive (family 0). 8 cells in a row seems to be a good start.
        for j in range(8):
            family_code[offset + j] = 0

    for i in range(n_init):
        offset = 0

        while (offset < 5 * num_row) or (offset > (num_row - 5) * num_row):
            offset = random.randint(0, num_cells-1)
        # Another interesting one is the spaceship pattern
        family_code[offset] = 0
        family_code[offset + 1] = 0
        family_code[offset + 2] = 0
        family_code[offset + 3] = 0
        family_code[offset + num_row - 1] = 0
        family_code[offset + num_row + 3] = 0
        family_code[offset + 2 * num_row + 3] = 0
        family_code[offset + 3 * num_row - 1] = 0
        family_code[offset + 3 * num_row + 2] = 0

    input_template_num += [template_sphere] * len(input_xyz)

    # All objects in the game are fixed
    DEMSim.SetFamilyFixed(0)
    DEMSim.SetFamilyFixed(10)

    # The force model just serves as a contact number register here. So let's say if 2 spheres are in contact, the
    # force is constant 1, so you have 4+ neighbours if you feel 4+ force, and 2- if you got 2- neighbours.
    # To avoid force cancelling out, we let alive cells always get positive force, and dead cells negative.
    model = DEMSim.DefineContactForceModel(
        "if (AOwnerFamily == 0) force = make_float3(0, 0, 1); else force = make_float3(0, 0, -1);")
    DEMSim.DisableContactBetweenFamilies(0, 0)
    DEMSim.DisableContactBetweenFamilies(10, 10)

    # The rule for changing family numbers is simple: you die, move to 10; live, move to 0. And then, family 10 will
    # not be outputted.
    # Dead to alive: if you have 3 alive neighbours
    DEMSim.ChangeFamilyWhen(
        10, 0, "float my_neighbours = length(acc * mass); return (my_neighbours > 2.9) && (my_neighbours < 3.1);")
    # Alive to dead, if less than 2 alive neighbours, or more than 3 alive neighbours (more than 6 dead neighbors, or
    # less than 5 dead neighbors)
    DEMSim.ChangeFamilyWhen(
        0, 10, "float my_neighbours = length(acc * mass); return (my_neighbours < 4.9) || (my_neighbours > 6.1);")

    particles = DEMSim.AddClumps(input_template_num, input_xyz)
    particles.SetFamilies(family_code)
    # The game board is somewhat large so we have to define it, because the solver defaults the world size to be 1000.
    DEMSim.InstructBoxDomainDimension(
        world_size * 1.1, world_size * 1.1, world_size * 1.1)
    DEMSim.SetInitBinSize(world_size / 100.)

    DEMSim.SetInitTimeStep(1.)
    DEMSim.SetCDUpdateFreq(0)
    # Must disable this if you want to run dT and kT synchronizely, or the solver will automatically find a non-zero
    # update frequency that it sees fit to run it in an async fashion.
    DEMSim.DisableAdaptiveUpdateFreq()
    DEMSim.Initialize()

    for i in range(3000):
        print(f"Outputting frame: {i}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{i:04d}.csv")
        DEMSim.WriteSphereFile(filename)
        print(
            f"Average contacts each sphere has: {DEMSim.GetAvgSphContacts()}", flush=True)

        DEMSim.DoDynamicsThenSync(1.)

    DEMSim.ShowMemStats()

    print("DEMdemo_GameOfLife exiting...")



# Copyright(c) 2021, SBEL GPU Development Team
# Copyright(c) 2021, University of Wisconsin - Madison
#
# SPDX - License - Identifier : BSD - 3 - Clause

# =============================================================
# A meshed ball hitting a granular bed under gravity.
# =============================================================

import DEME

import numpy as np
import os
import random

def force_model():
    model = """
        if (overlapDepth > 0.0) {
            // Material properties

            // DEM force calculation strategies for grain breakage
            float E_cnt, G_cnt, CoR_cnt, mu_cnt, Crr_cnt, E_A, E_B;
            {
                // E and nu are associated with each material, so obtain them this way
                E_A = E[bodyAMatType];
                float nu_A = nu[bodyAMatType];
                E_B = E[bodyBMatType];
                float nu_B = nu[bodyBMatType];
                matProxy2ContactParam<float>(E_cnt, G_cnt, E_A, nu_A, E_B, nu_B);
                // CoR, mu and Crr are pair-wise, so obtain them this way
                CoR_cnt = CoR[bodyAMatType][bodyBMatType];
                mu_cnt = mu[bodyAMatType][bodyBMatType];
                Crr_cnt = Crr[bodyAMatType][bodyBMatType];
            }
            float3 rotVelCPA, rotVelCPB;
            {
                // We also need the relative velocity between A and B in global frame to use in the damping terms
                // To get that, we need contact points' rotational velocity in GLOBAL frame
                // This is local rotational velocity (the portion of linear vel contributed by rotation)
                rotVelCPA = cross(ARotVel, locCPA);
                rotVelCPB = cross(BRotVel, locCPB);
                // This is mapping from local rotational velocity to global
                applyOriQToVector3<float, deme::oriQ_t>(rotVelCPA.x, rotVelCPA.y, rotVelCPA.z, AOriQ.w, AOriQ.x, AOriQ.y,
                                                        AOriQ.z);
                applyOriQToVector3<float, deme::oriQ_t>(rotVelCPB.x, rotVelCPB.y, rotVelCPB.z, BOriQ.w, BOriQ.x, BOriQ.y,
                                                        BOriQ.z);
            }
            float mass_eff, sqrt_Rd, beta;
            float3 vrel_tan;
            float3 delta_tan = make_float3(delta_tan_x, 0.0, delta_tan_z);

            // The (total) relative linear velocity of A relative to B
            const float3 velB2A = (ALinVel + rotVelCPA) - (BLinVel + rotVelCPB);
            const float projection = dot(velB2A, B2A);
            vrel_tan = velB2A - projection * B2A;
            vrel_tan.y = 0.0;

            const float3 v_rot = rotVelCPB - rotVelCPA;
            // This v_rot is only used for identifying resistance direction
            const float v_rot_mag = length(v_rot);
            mass_eff = (AOwnerMass * BOwnerMass) / (AOwnerMass + BOwnerMass);

            // Now we already have sufficient info to update contact history
            {
                delta_tan += ts * vrel_tan;
                const float disp_proj = dot(delta_tan, B2A);
                delta_tan -= disp_proj * B2A;
                delta_time += ts;
            }

            // Normal force part
            {
                sqrt_Rd = sqrt(overlapDepth * (ARadius * BRadius) / (ARadius + BRadius));
                const float Sn = 2. * E_cnt * sqrt_Rd;

                const float loge = (CoR_cnt < DEME_TINY_FLOAT) ? log(DEME_TINY_FLOAT) : log(CoR_cnt);
                beta = loge / sqrt(loge * loge + deme::PI_SQUARED);

                const float k_n = deme::TWO_OVER_THREE * Sn;
                const float gamma_n = deme::TWO_TIMES_SQRT_FIVE_OVER_SIX * beta * sqrt(Sn * mass_eff);

                force += (k_n * overlapDepth + gamma_n * projection) * B2A;
            }

            // Rolling resistance part
            if (Crr_cnt > 0.0) {
                // Figure out if we should apply rolling resistance force
                bool should_add_rolling_resistance = true;
                {
                    const float R_eff = sqrtf((ARadius * BRadius) / (ARadius + BRadius));
                    const float kn_simple = deme::FOUR_OVER_THREE * E_cnt * sqrtf(R_eff);
                    const float gn_simple = -2.f * sqrtf(deme::FIVE_OVER_THREE * mass_eff * E_cnt) * beta * powf(R_eff, 0.25f);

                    const float d_coeff = gn_simple / (2.f * sqrtf(kn_simple * mass_eff));

                    if (d_coeff < 1.0) {
                        float t_collision = deme::PI * sqrtf(mass_eff / (kn_simple * (1.f - d_coeff * d_coeff)));
                        if (delta_time <= t_collision) {
                            should_add_rolling_resistance = false;
                        }
                    }
                }
                // If should, then compute it (using Schwartz model)
                if (should_add_rolling_resistance) {
                    // Tangential velocity (only rolling contribution) of B relative to A, at contact point, in global
                    const float3 v_rot = make_float3(0.0, rotVelCPB.y - rotVelCPA.y, 0.0);
                    // This v_rot is only used for identifying resistance direction
                    const float v_rot_mag = length(v_rot);
                    if (v_rot_mag > DEME_TINY_FLOAT) {
                        // You should know that Crr_cnt * normal_force is the underlying formula, and in our model,
                        // it is a `force' that produces torque only, instead of also cancelling out friction.
                        // Its direction is that it `resists' rotation, see picture in
                        // https://en.wikipedia.org/wiki/Rolling_resistance.
                        torque_only_force = (v_rot / v_rot_mag) * (Crr_cnt * length(force));
                    }
                }
            }

            // Tangential force part
            if (mu_cnt > 0.0) {
                const float kt = 8. * G_cnt * sqrt_Rd;
                const float gt = -deme::TWO_TIMES_SQRT_FIVE_OVER_SIX * beta * sqrt(mass_eff * kt);
                float3 tangent_force = -kt * delta_tan - gt * vrel_tan;
                const float ft = length(tangent_force);
                if (ft > DEME_TINY_FLOAT) {
                    // Reverse-engineer to get tangential displacement
                    const float ft_max = length(force) * mu_cnt;
                    if (ft > ft_max) {
                        tangent_force = (ft_max / ft) * tangent_force;
                        delta_tan = (tangent_force + gt * vrel_tan) / (-kt);
                    }
                } else {
                    tangent_force = make_float3(0, 0, 0);
                }
                // Use force to collect tangent_force
                force += tangent_force;
            }

            // Make sure we update those wildcards (in this case, contact history)
            delta_tan_x = delta_tan.x;
            delta_tan_y = 0.0;
            delta_tan_z = delta_tan.z;
            force.y = 0.0;
        }
    """

    return model

if __name__ == "__main__":
    ball_density = 6.2e3
    H = 0.1
    R = 0.0254 / 2.0

    terrain_rad = 0.006 / 2.0

    out_dir = "DemoOutput_BallDrop2D/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("STEP_METRIC")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetOutputContent(["ABSV", "XYZ"])
    DEMSim.SetMeshOutputFormat("VTK")

    # E, nu, CoR, mu, Crr...
    mat_type_ball = DEMSim.LoadMaterial(
        {"E": 7e7, "nu": 0.24, "CoR": 0.9, "mu": 0.3, "Crr": 0.0}
    )
    mat_type_terrain = DEMSim.LoadMaterial(
        {"E": 7e7, "nu": 0.24, "CoR": 0.9, "mu": 0.3, "Crr": 0.0}
    )
    mat_type_terrain_sim = DEMSim.LoadMaterial(
        {"E": 7e7, "nu": 0.24, "CoR": 0.9, "mu": 0.3, "Crr": 0.0}
    )

    step_size = 2e-6
    world_size = 0.2
    DEMSim.InstructBoxDomainDimension(
        tuple([-world_size / 2.0, world_size / 2.0]),
        tuple([-world_size / 2.0, world_size / 2.0]),
        tuple([0, 10 * world_size]),
    )
    DEMSim.InstructBoxDomainBoundingBC("top_open", mat_type_terrain)

    projectile = DEMSim.AddWavefrontMeshObject(
        DEME.GetDEMEDataFile("mesh/sphere.obj"), mat_type_ball
    )
    projectile.Scale(R)
    print(f"Total num of triangles: {projectile.GetNumTriangles()}")

    projectile.SetInitPos([0, 0, 8 * world_size])
    ball_mass = ball_density * 4.0 / 3.0 * np.pi * R * R * R
    projectile.SetMass(ball_mass)
    projectile.SetMOI(
        [
            ball_mass * 2 / 5 * R * R,
            ball_mass * 2 / 5 * R * R,
            ball_mass * 2 / 5 * R * R,
        ]
    )
    projectile.SetFamily(2)
    DEMSim.SetFamilyFixed(2)
    DEMSim.DisableContactBetweenFamilies(0, 2)
    # Track the projectile
    proj_tracker = DEMSim.Track(projectile)

    # Force model to use
    # model2D = DEMSim.ReadContactForceModel("ForceModel2D.cu")
    model2D = DEMSim.DefineContactForceModel(force_model())
    model2D.SetMustHaveMatProp(set(["E", "nu", "CoR", "mu", "Crr"]))
    model2D.SetMustPairwiseMatProp(set(["CoR", "mu", "Crr"]))
    model2D.SetPerContactWildcards(
        set(["delta_time", "delta_tan_x", "delta_tan_y", "delta_tan_z"])
    )

    templates_terrain = []
    num_templates = 11
    for i in range(num_templates):
        templates_terrain.append(
            DEMSim.LoadSphereType(
                terrain_rad * terrain_rad * terrain_rad * 2.0e3 * 4 / 3 * np.pi,
                terrain_rad,
                mat_type_terrain,
            )
        )
        terrain_rad += 0.0001 / 2.0

    num_particle = 0
    sample_z = 1.5 * terrain_rad
    fullheight = world_size * 6.0
    sample_halfwidth = world_size / 2 - 2 * terrain_rad
    init_v = 0.01

    sampler = DEME.HCPSampler(2.05 * terrain_rad)

    sample_center = [0, 0, fullheight / 2 + 1 * terrain_rad]
    input_xyz = sampler.SampleBox(
        sample_center, [sample_halfwidth, 0.0, fullheight / 2.0]
    )
    template_to_use = []
    # Randomly select from pre - generated clump templates
    for i in range(len(input_xyz)):
        template_to_use.append(templates_terrain[random.randint(0, num_templates - 1)])

    DEMSim.AddClumps(template_to_use, input_xyz)
    num_particle += len(input_xyz)

    print(f"Total num of particles: {num_particle}")

    max_z_finder = DEMSim.CreateInspector("clump_max_z")
    total_mass_finder = DEMSim.CreateInspector("clump_mass")

    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetMaxVelocity(30.0)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])

    DEMSim.Initialize()

    sim_time = 3.0
    settle_time = 1.0
    fps = 20
    frame_time = 1.0 / fps
    out_steps = (int)(1.0 / (fps * step_size))

    print(f"Output at {fps} FPS")
    currframe = 0

    # We can let it settle first
    t = 0.0
    while t < settle_time:
        print(f"Frame: {currframe}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{currframe:04d}.csv")
        meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        currframe += 1

        DEMSim.DoDynamicsThenSync(frame_time)
        DEMSim.ShowThreadCollaborationStats()
        t += frame_time

    cp_filename = os.path.join(out_dir, f"bed.csv")
    DEMSim.WriteClumpFile(cp_filename)

    # This is to show that you can change the material for all the particles in a family... although here,
    # mat_type_terrain_sim and mat_type_terrain are the same material so there is no effect  you can define
    # them differently though.
    DEMSim.SetFamilyClumpMaterial(0, mat_type_terrain_sim)
    DEMSim.DoDynamicsThenSync(0.2)
    terrain_max_z = max_z_finder.GetValue()
    matter_mass = total_mass_finder.GetValue()
    total_volume = (world_size * world_size) * (terrain_max_z - 0.0)
    bulk_density = matter_mass / total_volume
    print(f"Original terrain height: {terrain_max_z}")
    print(f"Bulk density: {bulk_density}")

    # Then drop the ball
    DEMSim.ChangeFamily(2, 0)
    proj_tracker.SetPos([0, 0, terrain_max_z + R + H])

    t = 0.0
    while t < sim_time:
        print(f"Frame: {currframe}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{currframe:04d}.csv")
        meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        DEMSim.ShowThreadCollaborationStats()
        currframe += 1

        DEMSim.DoDynamicsThenSync(frame_time)
        DEMSim.ShowThreadCollaborationStats()
        t += frame_time

        if (abs(proj_tracker.Vel()[2]) < 1e-4) and (t > 0.1):
            break

        DEMSim.ShowTimingStats()

    DEMSim.ShowMemStats()

    final_pos = proj_tracker.Pos()
    print(f"Ball density: {ball_density}")
    print(f"Ball rad: {R}")
    print(f"Drop height: {H}")
    # final_pos[2] is the z coord
    print(f"Penetration: {terrain_max_z - (final_pos[2] - R)}")
    print(f"==============================================================")
    print(f"DEMdemo_BallDrop exiting...")



#  Copyright (c) 2021, SBEL GPU Development Team
#  Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# This demo shows how to control the deformation of a flexibile mesh and use
# it in a DEM simulation.
# We show how mesh node coordinates can be extracted and modified. This can
# potentially be used with an external solid mechanics solver to do co-simulations
# involving flexible bodies.
# =============================================================================

import DEME

import numpy as np
import os
import csv

math_PI = 3.1415927
force_csv_header = ["point_x", "point_y",
                    "point_z", "force_x", "force_y", "force_z"]

# Used to write pairwise force concerning the mesh to a file.


def writePointsForcesToCSV(force_csv_header,
                           points,
                           forces,
                           filename):
    combined_array = np.hstack((np.array(points), np.array(forces)))
    with open(filename, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, quoting=csv.QUOTE_NONE, escapechar=',')
        writer.writerow(force_csv_header)
        for row in combined_array:
            writer.writerow(row)


if __name__ == "__main__":
    out_dir = "DemoOutput_FlexibleMesh/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("INFO")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetOutputContent(["ABSV", "XYZ"])
    DEMSim.SetMeshOutputFormat("VTK")
    DEMSim.SetContactOutputContent(["OWNER", "FORCE", "POINT", "TORQUE"])

    # E, nu, CoR, mu, Crr...
    mat_type_mesh = DEMSim.LoadMaterial(
        {"E": 1e8, "nu": 0.3, "CoR": 0.5, "mu": 0.7, "Crr": 0.00})
    mat_type_particle = DEMSim.LoadMaterial(
        {"E": 1e8, "nu": 0.3, "CoR": 0.5, "mu": 0.4, "Crr": 0.00})
    # If you don't have this line, then values will take average between 2 materials, when they are in contact
    DEMSim.SetMaterialPropertyPair("mu", mat_type_mesh, mat_type_particle, 0.5)

    world_size = 2.
    # You have the choice to not specify the world size. In that case, the world is 20^3. And you might want to add
    # world boundaries yourself, like we did in this demo.
    DEMSim.InstructBoxDomainDimension([-world_size / 2, world_size / 2],
                                      [-world_size / 2, world_size / 2], [0, world_size])

    # No need to add simulation `world' boundaries, b/c we'll add a cylinderical container manually
    DEMSim.InstructBoxDomainBoundingBC("none", mat_type_mesh)
    # Now manually add boundaries (you can choose to add them automatically using InstructBoxDomainBoundingBC, too)
    walls = DEMSim.AddExternalObject()
    walls.AddPlane([0, 0, 0], [0, 0, 1], mat_type_mesh)
    walls.AddPlane([0, 0, world_size], [0, 0, -1], mat_type_mesh)
    walls.AddPlane([world_size / 2, 0, 0], [-1, 0, 0], mat_type_mesh)
    walls.AddPlane([-world_size / 2, 0, 0], [1, 0, 0], mat_type_mesh)
    walls.AddPlane([0, world_size / 2, 0], [0, -1, 0], mat_type_mesh)
    walls.AddPlane([0, -world_size / 2, 0], [0, 1, 0], mat_type_mesh)

    # Define the terrain particle templates
    # Calculate its mass and MOI
    terrain_density = 2.6e3
    clump_vol = 4. / 3. * math_PI
    mass = terrain_density * clump_vol
    MOI = np.array([2. / 5., 2. / 5., 2. / 5.]) * mass
    # Then load it to system
    my_template = DEMSim.LoadClumpType(mass, MOI.tolist(), DEME.GetDEMEDataFile(
        "clumps/spiky_sphere.csv"), mat_type_particle)
    my_template.SetVolume(clump_vol)
    # Decide the scalings of the templates we just created (so that they are... like particles, not rocks)
    scale = 0.05
    my_template.Scale(scale)

    # Sample 2 chunks of materials in this part
    sampler = DEME.HCPSampler(scale * 2.2)
    fill_height = 1.75
    fill_center1 = [-world_size / 4, 0, fill_height / 2 + 2 * scale]
    fill_center2 = [world_size / 4, 0, fill_height / 2 + 2 * scale]
    fill_halfsize = [world_size / 4 - 2 * scale,
                     world_size / 4 - 2 * scale, fill_height / 2]
    input_xyz1 = sampler.SampleBox(fill_center1, fill_halfsize)
    input_xyz2 = sampler.SampleBox(fill_center2, fill_halfsize)
    particles = DEMSim.AddClumps(my_template, input_xyz1 + input_xyz2)
    print(f"Total num of particles: {particles.GetNumClumps()}")
    print(f"Total num of spheres: {particles.GetNumSpheres()}")

    # Load in the mesh which is a 2x2 (yz) plate. Its thickness is 0.05 in the x direction.
    flex_mesh = DEMSim.AddWavefrontMeshObject(
        DEME.GetDEMEDataFile("mesh/thin_plate.obj"), mat_type_mesh, True, False)
    num_tri = flex_mesh.GetNumTriangles()
    print(f"Total num of triangles: {num_tri}")

    # The define the properties
    body_mass = 1.5e3 * 1. * 1. * 0.05
    flex_mesh.SetMass(body_mass)
    Iz = 1. / 12. * (0.05 * 0.05 + 1 * 1) * body_mass
    Ix = 1. / 12. * (1 * 1 + 1 * 1) * body_mass
    flex_mesh.SetMOI([Ix, Iz, Iz])
    # This mesh is created with CoM being 0,0,0, so no effect by InformCentroidPrincipal. But it is good that you know
    # this method exists. Also, [0,0,0,1] means unit quaternion instead of [1,0,0,0].
    flex_mesh.InformCentroidPrincipal([0, 0, 0], [0, 0, 0, 1])
    # Attach it to the ceiling
    flex_mesh.SetInitPos([0, 0, 1.2])
    flex_mesh.SetFamily(1)
    # If you call SetFamilyPrescribedPosition and SetFamilyPrescribedQuaternion without specifying what position it
    # actually take, then its position is kept `as is' during simulation, without being affected by physics. It's
    # similar to fixing it but allows you manually impose velocities (which may have implications on your force model),
    # even though the velocity won't change its location. If you prescribe position by do not prescribe velocities, it
    # may make the object accumulate `phantom' velocity and de-stabilize the simulation. Fixing both position and
    # velocity is equivalent to fixing the family.
    DEMSim.SetFamilyPrescribedPosition(1)
    DEMSim.SetFamilyPrescribedQuaternion(1)
    DEMSim.SetFamilyPrescribedLinVel(1)
    DEMSim.SetFamilyPrescribedAngVel(1)
    # DEMSim.SetFamilyFixed(1);

    # Track the mesh
    flex_mesh_tracker = DEMSim.Track(flex_mesh)

    # Some inspectors
    max_z_finder = DEMSim.CreateInspector("clump_max_z")

    step_size = 5e-6
    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])
    # Mesh has user-enforced deformation that the solver won't expect, so it can be better to allow larger safety
    # adder.
    DEMSim.SetExpandSafetyAdder(1.0)
    DEMSim.SetErrorOutAvgContacts(50)
    DEMSim.Initialize()

    # After system initialization, you can still get an handle of the mesh components using trackers (GetMesh method).
    # But note that if you do changes using this handle, the changes are done to the mesh and it immediate affects the
    # simulation. So sometimes you want to copy the information you are after and keep for your record.
    mesh_handle = flex_mesh_tracker.GetMesh()
    # This is keeping a copy of the RELATIVE (to the CoM) locations of the mesh nodes. In our case, the Z coordinates
    # of these nodes range from -0.5 to 0.5. In Python, you get a n by 3 matrix, and here we create a copy of this matrix.

    node_resting_location = np.array(mesh_handle.GetCoordsVertices())

    sim_end = 9.0
    fps = 20
    frame_time = 1.0 / fps
    print(f"Output at {fps} FPS")
    out_steps = (int)(1.0 / (fps * step_size))
    frame_count = 0
    step_count = 0

    # Used to store forces and points of contact
    forces = np.array([])
    points = np.array([])
    num_force_pairs = 0

    # Settle
    t = 0.
    while (t < 0.5):
        print(f"Outputting frame: {frame_count}", flush=True)
        filename = os.path.join(
            out_dir, f"DEMdemo_output_{frame_count:04d}.csv")
        meshname = os.path.join(
            out_dir, f"DEMdemo_mesh_{frame_count:04d}.vtk")
        force_filename = os.path.join(
            out_dir, f"DEMdemo_forces_{frame_count:04d}.csv")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        writePointsForcesToCSV(force_csv_header, points,
                               forces, force_filename)
        DEMSim.ShowThreadCollaborationStats()
        frame_count += 1

        DEMSim.DoDynamics(frame_time)
        t += frame_time

    # It's possible that you don't have to update the mesh every time step so you can set this number larger than 1.
    # However, you have to then ensure the simulation does not de-stabilize because the mesh--particles contacts are
    # running in a delayed fashion and large penetrations can occur. If the mesh is super soft, then it's probably OK.
    ts_per_mesh_update = 5
    # Some constants that are used to define the artificial mesh motion. You'll see in the main simulation loop.
    max_wave_magnitude = 0.3
    wave_period = 3.0

    # Main simulation loop starts...
    t = 0.
    while (t < sim_end):
        if (step_count % out_steps == 0):
            print(f"Outputting frame: {frame_count}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{frame_count:04d}.csv")
            meshname = os.path.join(
                out_dir, f"DEMdemo_mesh_{frame_count:04d}.vtk")
            force_filename = os.path.join(
                out_dir, f"DEMdemo_forces_{frame_count:04d}.csv")
            DEMSim.WriteSphereFile(filename)
            DEMSim.WriteMeshFile(meshname)
            frame_count += 1
            # We write force pairs that are related to the mesh to a file
            [points, forces] = flex_mesh_tracker.GetContactForces()
            writePointsForcesToCSV(
                force_csv_header, points, forces, force_filename)
            DEMSim.ShowThreadCollaborationStats()

        # We probably don't have to update the mesh every time step
        if (step_count % ts_per_mesh_update == 0):
            # For real use cases, you probably will use an external solver to solve the defomration of the mesh,
            # then feed it to DEME. Here, we create an artificial defomration pattern for the mesh based on mesh node
            # location and time. This is just for show.

            # First, get where the mesh nodes are currently. We copy so we can modify it later on.
            # Shlok: This GetMeshNodesGlobal method maps to cpp version GetMeshNodesGlobalAsVectorOfVector method.
            node_current_location = np.array(
                flex_mesh_tracker.GetMeshNodesGlobal())
            # If you need the current RELATIVE (to the CoM) locations of the mesh nodes instead of global coordinates,
            # you can get it like the following:
            # node_current_location = mesh_handle.GetCoordsVertices()

            # Now calculate how much each node should `wave' and update the node location array. Remember z = 1 is
            # where the highest (relative) mesh node is. Again, this is artificial and only for showcasing this
            # utility.
            for i in range(node_current_location.shape[0]):
                # Use resting locations to calculate the magnitude of waving for nodes...
                my_wave_distance = (
                    (1. - node_resting_location[i, 2]) / 2.)**2 * max_wave_magnitude * np.sin(t / wave_period * 2 * np.pi)
                # Then update the current location array...
                node_current_location[i,
                                      0] = node_resting_location[i, 0] + my_wave_distance

            # Now instruct the mesh to deform. Two things to pay attention to:

            # 1. We should respect the actual CoM location of the mesh. We get the global coords of mesh nodes using
            # GetMeshNodesGlobal, but UpdateMesh works with mesh's local or say relative coordinates, and that is why
            # we do FrameTransformGlobalToLocal first. And depending on your setup, the CoM and coord frame of
            # your mesh might be moving, and if it moves and rotates then you probably need to move and rotate the
            # points you got to offset the influence of CoM and local frame first. That said, if you use
            # mesh_handle.GetCoordsVertices() as I mentioned above to get the relative node positions of the mesh,
            # then no need to FrameTransformGlobalToLocal the CoM and rotate the frame.

            # 2. UpdateMesh will update the relative locations of mesh nodes to your specified locations. But if you
            # just have the information on the amount of mesh deformation, then you can use UpdateMeshByIncrement
            # instead, to incremenet mesh nodes' relative locations.

            mesh_CoM_pos = flex_mesh_tracker.Pos()
            mesh_frame_oriQ = flex_mesh_tracker.OriQ()
            for i in range(node_current_location.shape[0]):
                node_current_location[i, :] = DEME.FrameTransformGlobalToLocal(
                    node_current_location[i, :], mesh_CoM_pos, mesh_frame_oriQ)
            flex_mesh_tracker.UpdateMesh(node_current_location)

            # Forces need to be extracted, if you want to use an external solver to solve the mesh's deformation. You
            # can do it like shown below. In this example, we did not use it other than writing it to a file; however
            # you may want to feed the array directly to your soild mechanics solver.
            [points, forces] = flex_mesh_tracker.GetContactForces()

        # Means advance simulation by one time step
        DEMSim.DoStepDynamics()
        t += step_size
        step_count += 1

    DEMSim.ShowTimingStats()
    DEMSim.ShowMemStats()
    print("FlexibleMesh demo exiting...")



#  Copyright (c) 2021, SBEL GPU Development Team
#  Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# A simplified wheel drawbar-pull test, featuring Curiosity wheel geometry and
# terrain particles represented by irregular DEM elements. Unlike WheelDP, this
# demo does not have a prerequisite and can run on its own, and the terrain is
# simpler since there is only one type of particle (3-sphere clump). The velocity
# and angular velocity of the wheel are prescribed and the terrain force on the
# wheel is measured (we only test one slip case in this demo, to make it faster).
# =============================================================================

import DEME
from DEME import HCPSampler

import numpy as np
import os
# import ctypes

if __name__ == "__main__":
    out_dir = "DemoOutput_WheelDPSimplified/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("INFO")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetMeshOutputFormat("VTK")
    DEMSim.SetContactOutputContent(["OWNER", "FORCE", "POINT"])

    # If you don't need individual force information, then this option makes the solver run a bit faster.
    DEMSim.SetNoForceRecord(True)

    # E, nu, CoR, mu, Crr...
    mat_type_wheel = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.6, "mu": 0.5, "Crr": 0.01})
    mat_type_terrain = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.4, "mu": 0.5, "Crr": 0.01})

    # If you don't have this line, then mu between drum material and granular material will be the average of the two.
    DEMSim.SetMaterialPropertyPair("mu", mat_type_wheel, mat_type_terrain, 0.8)
    DEMSim.SetMaterialPropertyPair(
        "CoR", mat_type_wheel, mat_type_terrain, 0.6)

    # `World'
    G_mag = 9.81
    step_size = 5e-6
    world_size_y = 1.
    world_size_x = 2.
    world_size_z = 2.
    DEMSim.InstructBoxDomainDimension(
        world_size_x, world_size_y, world_size_z, "none")
    DEMSim.InstructBoxDomainBoundingBC("top_open", mat_type_terrain)
    bottom = -0.5
    bot_wall = DEMSim.AddBCPlane([0, 0, bottom], [0, 0, 1], mat_type_terrain)

    # Define the wheel geometry
    wheel_rad = 0.25
    wheel_width = 0.2
    wheel_weight = 0000100.000
    wheel_mass = wheel_weight / G_mag
    total_pressure = 200.0
    added_pressure = total_pressure - wheel_weight
    wheel_IYY = wheel_mass * wheel_rad * wheel_rad / 2
    wheel_IXX = (wheel_mass / 12) * (3 * wheel_rad *
                                     wheel_rad + wheel_width * wheel_width)

    wheel = DEMSim.AddWavefrontMeshObject(DEME.GetDEMEDataFile(
        "mesh/rover_wheels/viper_wheel_right.obj"), mat_type_wheel, True, False)

    wheel.SetMass(wheel_mass)
    wheel.SetMOI([wheel_IXX, wheel_IYY, wheel_IXX])

    # Give the wheel a family number so we can potentially add prescription
    wheel.SetFamily(1)

    # Define the terrain particle templates
    # Calculate its mass and MOI
    terrain_density = 2.6e3
    volume1 = 4.2520508
    mass1 = terrain_density * volume1
    MOI1 = np.array([1.6850426, 1.6375114, 2.1187753]) * terrain_density

    # Scale the template we just created
    scale = 0.02
    # Then load it to system
    my_template = DEMSim.LoadClumpType(mass1, MOI1.tolist(), DEME.GetDEMEDataFile(
        "clumps/triangular_flat.csv"), mat_type_terrain)
    # Now scale the template
    # Note the mass and MOI are also scaled in the process, automatically. But if you are not happy with this, you
    # can always manually change mass and MOI afterwards.
    my_template.Scale(scale)
    # Give these templates names, 0000, 0001 etc.
    template_num = 0
    t_name = f"{template_num:04d}"
    my_template.AssignName(t_name)

    # Sampler to use
    sampler = HCPSampler(scale * 2.7)
    sample_halfheight = 0.25
    sample_halfwidth_x = (world_size_x * 0.95) / 2
    sample_halfwidth_y = (world_size_y * 0.95) / 2
    offset_z = bottom + sample_halfheight + 0.03
    # Sample initial particles
    sample_center = np.array([0, 0, offset_z])
    terrain_particles_xyz = sampler.SampleBox(sample_center.tolist(
    ), [sample_halfwidth_x, sample_halfwidth_y, sample_halfheight])
    terrain_template_in_use = [my_template] * len(terrain_particles_xyz)
    heap_family = [0] * len(terrain_particles_xyz)

    terrain_particles = DEMSim.AddClumps(
        terrain_template_in_use, terrain_particles_xyz)
    # Give ground particles a small initial velocity so they `collapse' at the start of the simulation
    terrain_particles.SetVel([0.00, 0, -0.05])
    terrain_particles.SetFamilies(heap_family)
    print(f"Current number of clumps: {len(terrain_particles_xyz)}")

    # Track it
    bot_wall_tracker = DEMSim.Track(bot_wall)
    wheel_tracker = DEMSim.Track(wheel)

    # Families' prescribed motions
    math_PI = 3.1415927
    w_r = math_PI / 4
    v_ref = w_r * wheel_rad

    sim_end = 6.
    # Note: this wheel is not `dictated' by our prescrption of motion because it can still fall onto the ground
    # (move freely linearly)
    DEMSim.SetFamilyPrescribedAngVel(1, "0", f"{w_r:09}", "0", False)
    # An extra force (acceleration) is addedd to simulate the load that the wheel carries
    DEMSim.AddFamilyPrescribedAcc(
        1, "none", "none", f"{(-added_pressure / wheel_mass):09}")
    # `Real sim' family number
    DEMSim.SetFamilyPrescribedAngVel(2, "0", f"{w_r:09}", "0", False)
    # Note: this wheel is not `dictated' by our prescrption of motion (hence the false argument), because it
    # can sink into the ground (move on Z dir); but its X and Y motions are explicitly controlled.
    # This one says when the experiment is going, the slip ratio is 0.5 (by our prescribing linear and angular vel)
    DEMSim.SetFamilyPrescribedLinVel(
        2, f"{(v_ref * 0.5):09}", "0", "none", False)
    # An extra force (acceleration) is addedd to simulate the load that the wheel carries
    DEMSim.AddFamilyPrescribedAcc(
        2, "none", "none", f"{(-added_pressure / wheel_mass):09}")

    # Some inspectors
    max_z_finder = DEMSim.CreateInspector("clump_max_z")
    min_z_finder = DEMSim.CreateInspector("clump_min_z")
    total_mass_finder = DEMSim.CreateInspector("clump_mass")
    max_v_finder = DEMSim.CreateInspector("clump_max_absv")

    # Make ready for simulation
    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -G_mag])
    # Max velocity info is generally just for the solver's reference and the user do not have to set it. The solver
    # wouldn't take into account a vel larger than this when doing async-ed contact detection: but this vel won't
    # happen anyway and if it does, something already went wrong.
    DEMSim.SetMaxVelocity(20.)
    # Error out vel is used to force the simulation to abort when something goes wrong.
    DEMSim.SetErrorOutVelocity(35.)
    DEMSim.SetExpandSafetyMultiplier(1.)
    DEMSim.Initialize()

    fps = 10
    out_steps = (int)(1.0 / (fps * step_size))
    curr_step = 0
    currframe = 0
    frame_time = 1.0 / fps
    report_ps = 100
    report_steps = (int)(1.0 / (report_ps * step_size))
    print(f"Output at {fps} FPS")

    # Put the wheel in place, then let the wheel sink in initially
    max_z = max_z_finder.GetValue()
    wheel_tracker.SetPos([-0.45, 0, max_z + 0.03 + wheel_rad], 0)

    t = 0.
    while t < 1.:
        print(f"Outputting frame: {currframe}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{currframe:04d}.csv")
        meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        currframe += 1

        DEMSim.DoDynamics(frame_time)
        t += frame_time

    # Switch wheel from free fall into DP test
    DEMSim.DoDynamicsThenSync(0)
    DEMSim.ChangeFamily(1, 2)

    t = 0.
    while (t < sim_end):
        if (curr_step % out_steps == 0):
            print(f"Outputting frame: {currframe}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{currframe:04d}.csv")
            meshname = os.path.join(
                out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
            DEMSim.WriteSphereFile(filename)
            DEMSim.WriteMeshFile(meshname)
            currframe += 1
            DEMSim.ShowThreadCollaborationStats()

        if (curr_step % report_steps == 0):
            force_list = wheel_tracker.ContactAcc()
            forces = np.array(force_list) * wheel_mass
            print(f"Time: {t}", flush=True)
            print(
                f"Force on wheel: {forces[0]}, {forces[1]}, {forces[2]}", flush=True)
            print(
                f"Drawbar pull coeff: {(forces[0] / total_pressure)}", flush=True)

        DEMSim.DoDynamics(step_size)
        curr_step += 1
        t += step_size

    DEMSim.ShowTimingStats()
    DEMSim.ShowMemStats()

    print("WheelDPSimpilified demo exiting...")



#  Copyright (c) 2021, SBEL GPU Development Team
#  Copyright (c) 2021, University of Wisconsin - Madison
#
#  SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# This demo features an analytical boundary-represented fast rotating container
# with particles of various shapes pulled into it. Different types of particles
# are marked with different family numbers (identification numbers) for easier
# visualizations.
# =============================================================================

import DEME

import numpy as np
import os

if __name__ == "__main__":
    out_dir = "DemoOutput_Centrifuge/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetOutputFormat("CSV")
    # Output family numbers (used to identify the centrifuging effect)
    DEMSim.SetOutputContent(["ABSV", "FAMILY"])
    # DEMSim.SetVerbosity("STEP_METRIC");

    # If you don't need individual force information, then this option makes the solver run a bit faster.
    DEMSim.SetNoForceRecord(True)

    mat_type_sand = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.6, "mu": 0.5, "Crr": 0.01})
    mat_type_drum = DEMSim.LoadMaterial(
        {"E": 2e9, "nu": 0.3, "CoR": 0.6, "mu": 0.5, "Crr": 0.01})
    # Since two types of materials have the same mu, this following call does not change the default mu for their
    # interaction, it's still 0.5.
    DEMSim.SetMaterialPropertyPair("mu", mat_type_sand, mat_type_drum, 0.5)

    # We can scale this general template to make it smaller, like a DEM particle that you would actually use
    scaling = 0.01
    # Calculate its mass and MOI (scaled according to the size)
    mass = 2.6e3 * 4. / 3. * DEME.PI * 2 * 1 * 1
    MOI = np.array([1. / 5. * mass * (1 * 1 + 2 * 2), 1. / 5. *
                   mass * (1 * 1 + 2 * 2), 1. / 5. * mass * (1 * 1 + 1 * 1)])
    # What will be loaded from the file, is a template for ellipsoid with b = c = 1 and a = 2, where Z is the long axis
    ellipsoid = DEMSim.LoadClumpType(mass, MOI.tolist(), DEME.GetDEMEDataFile(
        "clumps/ellipsoid_2_1_1.csv"), mat_type_sand)
    # Scale the template we just created
    ellipsoid.Scale(scaling)
    mass *= scaling * scaling * scaling
    MOI *= scaling * scaling * scaling * scaling * scaling
    ellipsoid.SetMass(mass)
    ellipsoid.SetMOI(MOI.tolist())

    # Define material type for the particles (on a per-sphere-component basis)
    ellipsoid.SetMaterial(mat_type_sand)

    # Create some random clump templates for the filling materials
    # An array to store these generated clump templates
    clump_types = []
    # Then randomly create some clump templates for filling the drum
    for i in range(3):
        # A multiplier is added to the masses of different clumps, so that centrifuging separate those types. Consider
        # it separating materials with different densities.
        mult = 1.5**i
        # Then make a new copy of the template then do the scaling of mass
        ellipsoid_template = DEMSim.Duplicate(ellipsoid)
        ellipsoid_template.SetMass(mass*mult)
        ellipsoid_template.SetMOI((MOI*mult).tolist())

        # Load a (ellipsoid-shaped) clump and a sphere
        clump_types.append(ellipsoid_template)
        clump_types.append(DEMSim.LoadSphereType(
            mass*mult, np.cbrt(2.0) * scaling, mat_type_sand))

    # Add the centrifuge
    CylCenter = [0, 0, 0]
    CylAxis = [0, 0, 1]
    CylRad = 2.0
    CylHeight = 1.0
    CylMass = 1.0
    safe_delta = 0.03
    IZZ = CylMass * CylRad * CylRad / 2
    IYY = (CylMass / 12) * (3 * CylRad * CylRad + CylHeight * CylHeight)
    Drum = DEMSim.AddExternalObject()
    Drum.AddCylinder(CylCenter, CylAxis, CylRad, mat_type_drum, 0)
    Drum.SetMass(CylMass)
    Drum.SetMOI([IYY, IYY, IZZ])
    Drum_tracker = DEMSim.Track(Drum)
    # Drum is family 100
    drum_family = 100
    Drum.SetFamily(drum_family)
    # The drum rotates (facing Z direction)
    DEMSim.SetFamilyPrescribedAngVel(drum_family, "0", "0", "6.0")
    # Then add planes to `close up' the drum. We add it as another object b/c we want to track the force on it
    # separately.
    top_bot_planes = DEMSim.AddExternalObject()
    top_bot_planes.AddPlane(
        [0, 0, CylHeight / 2. - safe_delta], [0, 0, -1], mat_type_drum)
    top_bot_planes.AddPlane(
        [0, 0, -CylHeight / 2. + safe_delta], [0, 0, 1], mat_type_drum)
    # Planes should rotate together with the drum wall.
    top_bot_planes.SetFamily(drum_family)
    planes_tracker = DEMSim.Track(top_bot_planes)

    # Then sample some particles inside the drum
    input_template_type = []
    family_code = []
    sample_center = [0, 0, 0]
    sample_halfheight = CylHeight / 2.0 - 3.0 * safe_delta
    sample_halfwidth = CylRad / 1.5
    input_xyz = DEME.DEMBoxGridSampler(sample_center, [sample_halfwidth, sample_halfwidth, sample_halfheight],
                                       scaling * np.cbrt(2.0) * 2.1, scaling * np.cbrt(2.0) * 2.1, scaling * 2 * 2.1)
    num_clumps = len(input_xyz)
    # Casually select from generated clump types
    for i in range(num_clumps):
        input_template_type.append(clump_types[i % len(clump_types)])
        # Every clump type that has a unique mass, gets a unique family number
        family_code.append((i % len(clump_types)) // 2)

    particles = DEMSim.AddClumps(input_template_type, input_xyz)
    particles.SetFamilies(family_code)

    # Keep tab of the max velocity in simulation
    max_v_finder = DEMSim.CreateInspector("clump_max_absv")
    max_v = -1.

    # Make the domain large enough
    DEMSim.InstructBoxDomainDimension(5, 5, 5)
    step_size = 5e-6
    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])
    DEMSim.SetExpandSafetyType("auto")
    # If there is a velocity that an analytical object (i.e. the drum) has that you'd like the solver to take into
    # account in consideration of adding contact margins, you have to specify it here, since the solver's automatic max
    # velocity derivation algorithm currently cannot take analytical object's angular velocity-induced velocity into
    # account.
    DEMSim.SetExpandSafetyAdder(6.0)
    DEMSim.Initialize()

    time_end = 20.0
    fps = 20
    out_steps = (int)(1.0 / (fps * step_size))

    print(f"Output at {fps} FPS")
    currframe = 0
    curr_step = 0
    t = 0.
    while (t < time_end):
        if (curr_step % out_steps == 0):
            print(f"Frame: {currframe}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{currframe:04d}.csv")
            DEMSim.WriteSphereFile(filename)
            currframe += 1

            DEMSim.ShowThreadCollaborationStats()
            max_v = max_v_finder.GetValue()
            print(
                f"Max velocity of any point in simulation is {max_v}", flush=True)

            # Torque on the side walls are?
            drum_moi = np.array(Drum_tracker.MOI())
            drum_pos = np.array(Drum_tracker.ContactAngAccLocal())
            drum_torque = np.multiply(drum_pos, drum_moi)
            print(
                f"Contact torque on the side walls is {drum_torque[0]}, {drum_torque[1]}, {drum_torque[2]}", flush=True)

            # The force on the bottom plane?
            force_on_BC = np.array(
                planes_tracker.ContactAcc()) * planes_tracker.Mass()
            print(
                f"Contact force on bottom plane is {force_on_BC[2]}", flush=True)

        DEMSim.DoDynamics(step_size)
        t += step_size
        curr_step += 1

    DEMSim.ShowThreadCollaborationStats()
    DEMSim.ClearThreadCollaborationStats()

    DEMSim.ShowTimingStats()
    DEMSim.ShowMemStats()

    print("DEMdemo_Centrifuge exiting...")



# Copyright (c) 2021, SBEL GPU Development Team
# Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

# =============================================================================
# A meshed ball hitting a granular bed under gravity.
# ========= ====================================================================

import DEME

import numpy as np
import os

if __name__ == "__main__":
    out_dir = "DemoOutput_BallDrop/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("STEP_METRIC")
    DEMSim.SetOutputFormat("CSV")
    DEMSim.SetOutputContent(["ABSV", "XYZ"])
    DEMSim.SetMeshOutputFormat("VTK")
    DEMSim.EnsureKernelErrMsgLineNum(True)

    # E, nu, CoR, mu, Crr... Material properties
    mat_type_ball = DEMSim.LoadMaterial(
        {"E": 1e10, "nu":  0.3, "CoR":  0.6, "mu":  0.3, "Crr": 0.01})
    mat_type_terrain = DEMSim.LoadMaterial(
        {"E":  5e9, "nu":  0.3, "CoR":  0.8, "mu":  0.3, "Crr":  0.01})
    # If you don't have this line, then CoR between mixer material and granular material will be 0.7 (average of the
    # two).
    DEMSim.SetMaterialPropertyPair("CoR", mat_type_ball, mat_type_terrain, 0.6)
    # Should do the same for mu and Crr, but since they are the same across 2 materials, it won't have an effect...

    step_size = 1e-5
    world_size = 10

    DEMSim.InstructBoxDomainDimension(
        tuple([0, world_size]), tuple([0, world_size]), tuple([0, world_size]), "none")
    DEMSim.InstructBoxDomainBoundingBC("top_open", mat_type_terrain)

    projectile = DEMSim.AddWavefrontMeshObject(
        DEME.GetDEMEDataFile("mesh/sphere.obj"), mat_type_ball, True, False)
    print(f"Total num of triangles: {projectile.GetNumTriangles()}")

    projectile.SetInitPos([world_size / 2, world_size / 2, world_size / 3 * 2])
    ball_mass = 7.8e3 * 4 / 3 * 3.1416
    projectile.SetMass(ball_mass)
    projectile.SetMOI(
        [ball_mass * 2 / 5, ball_mass * 2 / 5, ball_mass * 2 / 5])
    projectile.SetFamily(2)
    DEMSim.SetFamilyFixed(2)

    terrain_rad = 0.05
    template_terrain = DEMSim.LoadSphereType(
        terrain_rad * terrain_rad * terrain_rad * 2.6e3 * 4 / 3 * 3.14,  terrain_rad, mat_type_terrain)

    # Track the projectile
    proj_tracker = DEMSim.Track(projectile)

    sample_halfheight = world_size / 8
    sample_center = [world_size / 2, world_size / 2, sample_halfheight + 0.05]
    sample_halfwidth = world_size / 2 * 0.95

    input_xyz = DEME.DEMBoxHCPSampler(sample_center, [
                                      sample_halfwidth, sample_halfwidth, sample_halfheight], 2.01 * terrain_rad)
    DEMSim.AddClumps(template_terrain, input_xyz)
    print(f"Total num of particles: {len(input_xyz)}")

    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])
    # Max velocity info is generally just for the solver's reference and the user do not have to set it. The solver
    # wouldn't take into account a vel larger than this when doing async-ed contact detection: but this vel won't
    # happen anyway and if it does, something already went wrong.
    DEMSim.SetMaxVelocity(15.)
    # In general you don't have to worry about SetExpandSafetyAdder, unless if an entity has the property that a point
    # on it can move much faster than its CoM. In this demo, you are dealing with a meshed ball and you in fact don't
    # have this problem. In the Centrifuge demo though, this can be a problem since the centrifuge's CoM is not moving,
    # but its pointwise velocity can be high, so it needs to be accounted for using this method.
    DEMSim.SetExpandSafetyAdder(5.)
    DEMSim.SetInitBinSize(4 * terrain_rad)
    DEMSim.Initialize()

    sim_time = 6.0
    settle_time = 2.0
    fps = 20
    frame_time = (1.0 / fps)

    print(f"Output at {fps} FPS")
    currframe = 0

    # We can let it settle first
    t = 0.
    while (t < settle_time):
        print(f"Frame: {currframe}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{currframe:04d}.csv")
        meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        currframe += 1

        DEMSim.DoDynamicsThenSync(frame_time)
        DEMSim.ShowThreadCollaborationStats()
        t += frame_time

    # Then drop the ball. I also wanted to test if changing step size method works fine here...
    step_size *= 0.5
    DEMSim.UpdateStepSize(step_size)
    DEMSim.ChangeFamily(2, 1)

    t = 0.
    while (t < sim_time):
        print(f"Frame: {currframe}", flush=True)
        filename = os.path.join(out_dir, f"DEMdemo_output_{currframe:04d}.csv")
        meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
        DEMSim.WriteSphereFile(filename)
        DEMSim.WriteMeshFile(meshname)
        DEMSim.ShowThreadCollaborationStats()
        currframe += 1

        DEMSim.DoDynamicsThenSync(frame_time)
        DEMSim.ShowThreadCollaborationStats()
        t += frame_time

    DEMSim.ShowTimingStats()
    DEMSim.ShowMemStats()
    print("BallDrop demo exiting...")



#  Copyright (c) 2021, SBEL GPU Development Team
#  Copyright (c) 2021, University of Wisconsin - Madison
#
# SPDX-License-Identifier: BSD-3-Clause

import DEME
from DEME import HCPSampler

import numpy as np
import os

# =============================================================================
# In GRCPrep demo series, we try to prepare a sample of the GRC simulant, which
# are supposed to be used for extraterrestrial rover mobility simulations. It is
# made of particles of various sizes and shapes following a certain distribution.
# In Part1, it creates several batches of clumps and let them settle at the bottom
# of the domain.
# =============================================================================

if __name__ == "__main__":
    out_dir = "DemoOutput_GRCPrep_Part1/"
    out_dir = os.path.join(os.getcwd(), out_dir)
    os.makedirs(out_dir, exist_ok=True)

    DEMSim = DEME.DEMSolver()
    DEMSim.SetVerbosity("INFO")
    DEMSim.SetOutputFormat("CSV")
    # XYZ is default so this doesn't do anything
    DEMSim.SetOutputContent(["XYZ"])

    # Define materials
    mat_type_terrain = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.3, "mu": 0.5})
    mat_type_wheel = DEMSim.LoadMaterial(
        {"E": 1e9, "nu": 0.3, "CoR": 0.3, "mu": 0.5})

    # Define the simulation world
    world_y_size = 0.99
    DEMSim.InstructBoxDomainDimension(world_y_size, world_y_size, world_y_size)
    # Add 5 bounding planes around the simulation world, and leave the top open
    DEMSim.InstructBoxDomainBoundingBC("top_open", mat_type_terrain)
    bottom = -0.5
    DEMSim.AddBCPlane([0, 0, bottom], [0, 0, 1], mat_type_terrain)

    # Define the terrain particle templates
    # Calculate its mass and MOI
    terrain_density = 2.6e3
    volume1 = 4.2520508
    mass1 = terrain_density * volume1
    MOI1 = np.array([1.6850426, 1.6375114, 2.1187753]) * terrain_density
    volume2 = 2.1670011
    mass2 = terrain_density * volume2
    MOI2 = np.array([0.57402126, 0.60616378, 0.92890173]) * terrain_density
    # Scale the template we just created
    scales = [0.014, 0.0075833, 0.0044, 0.003, 0.002, 0.0018333, 0.0017]
    # Then load it to system
    my_template2 = DEMSim.LoadClumpType(mass2, MOI2.tolist(), DEME.GetDEMEDataFile(
        "clumps/triangular_flat_6comp.csv"), mat_type_terrain)
    my_template1 = DEMSim.LoadClumpType(mass1, MOI1.tolist(), DEME.GetDEMEDataFile(
        "clumps/triangular_flat.csv"), mat_type_terrain)
    ground_particle_templates = [my_template2, DEMSim.Duplicate(my_template2),
                                 my_template1,
                                 DEMSim.Duplicate(my_template1),
                                 DEMSim.Duplicate(my_template1),
                                 DEMSim.Duplicate(my_template1),
                                 DEMSim.Duplicate(my_template1)]
    # Now scale those templates
    for i in range(len(scales)):
        my_template = ground_particle_templates[i]
        # Note the mass and MOI are also scaled in the process, automatically. But if you are not happy with this, you
        # can always manually change mass and MOI afterwards.
        my_template.Scale(scales[i])
        # Give these templates names, 0000, 0001 etc.
        t_name = f"{i:04d}"
        my_template.AssignName(t_name)

    # Instatiate particles with a probability that is in line with their weight distribution.
    weight_perc = [0.1700, 0.2100, 0.1400, 0.1900, 0.1600, 0.0500, 0.0800]
    grain_perc = []
    for i in range(len(scales)):
        grain_perc.append(weight_perc[i] / (scales[i]**3))

    tmp = sum(grain_perc)
    grain_perc = [x / tmp for x in grain_perc]

    # Sampler to use
    sampler = HCPSampler(scales[0] * 2.2)

    # Make ready for simulation
    step_size = 1e-6
    DEMSim.SetInitTimeStep(step_size)
    DEMSim.SetGravitationalAcceleration([0, 0, -9.81])
    # Max velocity info is generally just for the solver's reference and the user do not have to set it. The solver
    # wouldn't take into account a vel larger than this when doing async-ed contact detection: but this vel won't
    # happen anyway and if it does, something already went wrong.
    DEMSim.SetMaxVelocity(15.)
    # Error out vel is used to force the simulation to abort when something goes wrong.
    DEMSim.SetErrorOutVelocity(15.)
    DEMSim.SetExpandSafetyMultiplier(1.2)
    DEMSim.SetInitBinSize(scales[2])
    DEMSim.Initialize()

    time_end = 10.0

    currframe = 0
    curr_step = 0

    sample_halfheight = 0.4
    sample_halfwidth_x = (world_y_size * 0.96) / 2
    sample_halfwidth_y = (world_y_size * 0.96) / 2
    offset_z = bottom + sample_halfheight + 0.15
    settle_frame_time = 0.2
    settle_batch_time = 2.0

    while (DEMSim.GetNumClumps() < 0.25e6):
        # DEMSim.ClearCache() # Clearing cache is no longer needed
        sample_center = [0, 0, offset_z]

        # Sample and add heap particles
        heap_particles_xyz = sampler.SampleBox(
            sample_center, [sample_halfwidth_x, sample_halfwidth_y, sample_halfheight])

        # Pick from possible grain templates with a pre-defined possibility
        heap_family = np.random.choice(
            range(len(grain_perc)), len(heap_particles_xyz), p=grain_perc)
        heap_template_in_use = [ground_particle_templates[i]
                                for i in heap_family]

        heap_particles = DEMSim.AddClumps(
            heap_template_in_use, heap_particles_xyz)
        # Give ground particles a small initial velocity so they `collapse' at the start of the simulation
        heap_particles.SetVel([0.00, 0, -0.05])
        heap_particles.SetFamilies(heap_family)
        DEMSim.UpdateClumps()
        print(f"Current number of clumps: {DEMSim.GetNumClumps()}")

        # Allow for some settling
        # Must DoDynamicsThenSync (not DoDynamics), as adding entities to the simulation is only allowed at a sync-ed
        # point of time.
        t = 0.
        while (t < settle_batch_time):
            print(f"Frame: {currframe}", flush=True)
            filename = os.path.join(
                out_dir, f"DEMdemo_output_{currframe:04d}.csv")
            # meshname = os.path.join(out_dir, f"DEMdemo_mesh_{currframe:04d}.vtk")
            DEMSim.WriteSphereFile(filename)
            # DEMSim.WriteMeshFile(meshname)
            currframe += 1

            DEMSim.DoDynamicsThenSync(settle_frame_time)
            t += settle_frame_time

        DEMSim.ShowThreadCollaborationStats()

    # Settle for some time more
    DEMSim.DoDynamicsThenSync(1.0)

    DEMSim.ShowThreadCollaborationStats()
    DEMSim.ClearThreadCollaborationStats()
    DEMSim.ShowMemStats()

    cp_filename = os.path.join(out_dir, f"GRC_3e5.csv")
    cnt_filename = os.path.join(out_dir, f"Contact_pairs_3e5.csv")
    DEMSim.WriteClumpFile(cp_filename)
    DEMSim.WriteMeshFile(cnt_filename)

    print("DEMdemo_GRCPrep_Part1 exiting...")
